import { reduce, flatMap, map, range, flatten, sampleSize } from 'lodash';

// import reduce from 'lodash/reduce'
// import flatMap from 'lodash/flatMap'
// import map from 'lodash/map'
// import range from 'lodash/range'


function assert(cond, msg) {
    if (!cond) {
        throw new Error(msg);
    }
}

function cartesianProduct(...args) {
    return reduce(args, (acc, val) => flatMap(acc, a => map(val, v => [].concat(a, v))));
}

const cellStates = {
    closed: 'closed',
    open: 'open',
    marked: 'marked'
    // legal transitions:
    //  closed --> marked (doMark)
    //  marked --> closed (doMark)
    //  closed --> open (doOpen)
};

class Cell {

    constructor({ state, mined, x, y, bombsAround = 0 }) {
        this.state = state;
        this.mined = mined;
        this.x = x;
        this.y = y;
        this.bombsAround = bombsAround;
    }

    config() {
        const { state, mined, x, y, bombsAround } = this;
        return { state, mined, x, y, bombsAround };
    }

    doMark() {
        if (this.state === cellStates.closed) {
            // do mark
            this.state = cellStates.marked;
        } else if (this.state === cellStates.marked) {
            // do unmark
            this.state = cellStates.closed;
        } else {
            throw new Error('doMark: Invalid state: ' + this.state);
        }
    }
    doOpen() {
        assert(this.state === cellStates.closed, 'doOpen: Invalid state: ' + this.state);
        this.state = cellStates.open;
    }
}

const openResults = {
    boom: 'boom',
    empty: 'empty',
    noop: 'noop'
};

const markResults = {
    ok: 'ok',
    noop: 'noop'
};

class MineField {

    constructor(size) {
        this.size = size;
        this.reset();
    }
    reset() {
        const [xAxis, yAxis] = [range(this.size), range(this.size)];
        this.grid = xAxis.map(x => yAxis.map(y => new Cell({ state: cellStates.closed, mined: false, x, y })));
    }
    initialize(bombs) {
        this.bombs = bombs;
        this.bombsCount = bombs.length;
        bombs.forEach(({ x, y }) => this.grid[x][y].mined = true);
        flatten(this.grid).forEach(cell => cell.bombsAround = this.bombsAround({ x: cell.x, y: cell.y }));
    }
    initializeRandom(bombsCount) {
        const bombs = sampleSize(cartesianProduct(range(this.size), range(this.size)), bombsCount).map(([x, y]) => ({ x, y }));
        this.initialize(bombs);
    }
    open(cell) {
        if (cell.state !== cellStates.closed) {
            return openResults.noop;
        }
        cell.doOpen();
        if (cell.mined) {
            return openResults.boom;
        }

        if (cell.bombsAround === 0) {
            nearbyCells({ x: cell.x, y: cell.y }, this.grid).filter(cell => cell.mined === false && cell.state === cellStates.closed).forEach(cell => this.open(cell));
        }
        return openResults.empty;
    }
    openAll() {
        flatten(this.grid).forEach(cell => this.open(cell));
    }
    remainingBombs() {
        return flatten(this.grid).filter(cell => cell.mined && cell.state !== cellStates.marked);
    }
    markedCells() {
        return flatten(this.grid).filter(cell => cell.state === cellStates.marked);
    }
    done() {
        return this.remainingBombs().length === 0;
    }
    mark(cell) {
        try {
            cell.doMark();
        } catch (e) {
            console.warn(e);
            return markResults.noop;
        }
        return markResults.ok;
    }
    bombsAround({ x, y }) {
        return nearbyCells({ x, y }, this.grid).filter(cell => cell.mined).length;
    }
}

function freeze(field) {
    const { size, bombsCount, bombs, grid } = field;
    const params = {
        size, bombsCount, bombs, grid: grid.map(row => row.map(cell => cell.config()))
    };
    return JSON.stringify(params, null, 4);
}

function revive(json) {
    const params = JSON.parse(json);
    const field = new MineField(params.size);
    field.bombs = params.bombs;
    field.bombsCount = params.bombsCount;
    field.grid = params.grid.map(row => row.map(cellConfig => new Cell(cellConfig)));
    return field;
}

// Nearby cells of the square field.
function nearbyCells({ x, y }, grid) {
    const diff = [0, 1, -1];
    return cartesianProduct(diff, diff).filter(([dx, dy]) => dx !== 0 || dy !== 0).map(([dx, dy]) => ({ x: x + dx, y: y + dy })).filter(({ x, y }) => [x, y].every(p => p >= 0 && p < grid.length)).map(({ x, y }) => grid[x][y]);
}

export { openResults, markResults, MineField, freeze, revive, cellStates, Cell };
