//@flow

// import range from 'lodash/range'
// import flatten from 'lodash/flatten'
import {flatten, range, sampleSize} from 'lodash'
import {Cell, cellStates} from './cell'
import {cartesianProduct} from './helpers'

import type {CellConfig} from './cell'


export type Point = {x: number, y: number}

export const openResults = {
    boom: 'boom',
    empty: 'empty',
    noop: 'noop'
}

export type OpenResult = $Keys<typeof openResults>

export const markResults = {
    ok: 'ok',
    noop: 'noop'
}

export type MarkResult = $Keys<typeof markResults>

export class MineField {
    size: number
    bombsCount: number
    grid: Array<Array<Cell>>
    bombs: Array<Point>

    constructor(size: number) {
        this.size = size
        this.reset()
    }
    reset() {
        const [xAxis, yAxis] = [range(this.size), range(this.size)]
        this.grid = xAxis.map(x =>
            yAxis.map(y => new Cell({state: cellStates.closed, mined: false, x, y}))
        )
    }
    initialize(bombs: Array<Point>) {
        this.bombs = bombs
        this.bombsCount = bombs.length
        bombs.forEach(({x, y}) => this.grid[x][y].mined = true)
        flatten(this.grid).forEach(cell => cell.bombsAround = this.bombsAround({x: cell.x, y: cell.y}))
    }
    initializeRandom(bombsCount: number) {
        const bombs = sampleSize(cartesianProduct(range(this.size), range(this.size)), bombsCount)
            .map(([x, y]) => ({x, y}))
        this.initialize(bombs)
    }
    open(cell: Cell): OpenResult {
        if (cell.state !== cellStates.closed) {
            return openResults.noop
        }
        cell.doOpen()
        if (cell.mined) {
            return openResults.boom
        }

        if (cell.bombsAround === 0) {
            nearbyCells({x: cell.x, y: cell.y}, this.grid)
                .filter(cell => cell.mined === false && cell.state === cellStates.closed)
                .forEach(cell => this.open(cell))
        }
        return openResults.empty
    }
    openAll() {
        flatten(this.grid).forEach(cell => this.open(cell))
    }
    remainingBombs() {
        return flatten(this.grid)
            .filter(cell => cell.mined && cell.state !== cellStates.marked)
    }
    markedCells() {
        return flatten(this.grid)
            .filter(cell => cell.state === cellStates.marked)
    }
    done() {
        return this.remainingBombs().length === 0
    }
    mark(cell: Cell): MarkResult {
        try {
            cell.doMark()
        } catch (e) {
            console.warn(e)
            return markResults.noop
        }
        return markResults.ok
    }
    bombsAround({x, y}: Point) {
        return nearbyCells({x, y}, this.grid).filter(cell => cell.mined).length
    }
}

type Params = {
    bombs: Array<Point>,
    bombsCount: number,
    grid: Array<Array<CellConfig>>,
    size: number
}

export function freeze(field: MineField): string {
    const {size, bombsCount, bombs, grid} = field
    const params: Params = {
        size, bombsCount, bombs, grid: grid.map(row => row.map(cell => cell.config()))
    }
    return JSON.stringify(params, null, 4)
}

export function revive(json: string): MineField {
    const params: Params = JSON.parse(json)
    const field = new MineField(params.size)
    field.bombs = params.bombs
    field.bombsCount = params.bombsCount
    field.grid = params.grid.map(row => row.map(cellConfig => new Cell(cellConfig)))
    return field
}


// Nearby cells of the square field.
function nearbyCells({x, y}: Point, grid: Array<Array<Cell>>): Array<Cell> {
    const diff = [0, 1, -1]
    return cartesianProduct(diff, diff)
        .filter(([dx, dy]) => dx !== 0 || dy !== 0)
        .map(([dx, dy]) => ({x: x + dx, y: y + dy}))
        .filter(({x, y}) => [x, y].every(p => p >= 0 && p < grid.length))
        .map(({x, y}) => grid[x][y])
}